import asyncio

from openai import AsyncStream
from openai.types.chat import ChatCompletionChunk

from .frontend import AsyncJsonNode
from .parser import AsyncParser
from .store import PathStore


def jsontap(s: AsyncStream[ChatCompletionChunk]):
    store = PathStore()
    parser = AsyncParser(store)

    async def feed():
        async for chunk in s:
            if chunk.choices and (delta := chunk.choices[0].delta) and delta.content:
                parser.feed(delta.content)

    asyncio.create_task(feed())
    asyncio.create_task(parser.parse())
    return AsyncJsonNode((), store)


"""
async def main():
    SCHEMA = {
        "type": "object",
        "properties": {
            "user": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "age": {"type": "integer"},
                    "scores": {"type": "array", "items": {"type": "integer"}},
                    "children": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "name": {"type": "string"},
                                "age": {"type": "integer"},
                            },
                            "required": ["name", "age"],
                            "additionalProperties": False,
                        },
                    },
                    "address": {
                        "type": "object",
                        "properties": {
                            "city": {"type": "string"},
                            "state": {"type": "string"},
                            "zip": {"type": "string"},
                        },
                        "required": ["city", "state", "zip"],
                        "additionalProperties": False,
                    },
                },
                "required": ["name", "age", "scores", "children", "address"],
                "additionalProperties": False,
            },
        },
        "required": ["user"],
        "additionalProperties": False,
    }
    client = AsyncOpenAI(
        api_key=os.environ.get("OPENAI_API_KEY")
        or os.environ.get("OPENROUTER_API_KEY"),
        base_url=os.environ.get("OPENAI_BASE_URL", "https://openrouter.ai/api/v1"),
    )

    stream = await client.chat.completions.create(
        model="nvidia/nemotron-3-nano-30b-a3b:free",
        messages=[
            {
                "role": "system",
                "content": (
                    "Generate a realistic user profile matching the provided JSON schema. "
                    "Include at least 3 scores and at least 2 children. Respond only with valid JSON."
                ),
            },
        ],
        response_format={
            "type": "json_schema",
            "json_schema": {
                "name": "user_profile",
                "strict": True,
                "schema": SCHEMA,
            },
        },
        stream=True,
    )

    json = jsontap(stream)
    async for child in json["user"]["children"]:
        print(f"  child: {await child['name']}, age {await child['age']}")

    print("name:", await json["user"]["name"])
    print("age:", await json["user"]["age"])
    print("scores:", await json["user"]["scores"])
    print("city:", await json["user"]["address"]["city"])
    print("state:", await json["user"]["address"]["state"])
    print("zip:", await json["user"]["address"]["zip"])
    print("user:", await json["user"])


if __name__ == "__main__":
    asyncio.run(main())
"""



from typing import Any

import asyncio
import ijson

from .store import PathStore


class AsyncParser:
    def __init__(self, store: PathStore):
        self._events = ijson.sendable_list()
        self._coro = ijson.parse_coro(self._events)
        self._queue = asyncio.Queue()
        self._store = store

    async def parse(self):
        return await self.parse_value(())

    async def _next_event(self) -> tuple[str, str, Any]:
        return await self._queue.get()

    def feed(self, chunk: str) -> None:
        self._coro.send(chunk.encode("utf-8"))
        for e in self._events:
            self._queue.put_nowait(e)
        del self._events[:]

    async def parse_value(self, prefix: tuple[str | int, ...]) -> Any:
        _, event, value = await self._next_event()
        if event == "start_map":
            node = await self.parse_object(prefix)
            return node
        elif event == "start_array":
            node = await self.parse_array(prefix)
            return node
        else:
            self._store.set(prefix, value)
            return value

    async def parse_object(self, prefix: tuple[str | int, ...]) -> dict[str | int, Any]:
        obj = {}
        while True:
            _, event, value = await self._next_event()
            if event == "map_key":
                obj[value] = await self.parse_value((*prefix, value))
            elif event == "end_map":
                break

        self._store.set(prefix, obj)
        return obj

    async def parse_array(self, prefix: tuple[str | int, ...]) -> list[Any]:
        arr = []
        index = 0
        while True:
            _, event, value = await self._next_event()
            if event == "end_array":
                break
            self._store.begin_item(prefix)
            if event == "start_map":
                arr.append(await self.parse_object((*prefix, index)))
                index += 1
            elif event == "start_array":
                arr.append(await self.parse_array((*prefix, index)))
                index += 1
            else:
                arr.append(value)
                self._store.set((*prefix, index), value)
                index += 1

        self._store.set(prefix, arr)
        return arr



import asyncio
from typing import Any
from collections import defaultdict
from attrs import define, field


"""
for arrays,

when a user does

async for x in j["arr"]:

__aiter__ is called
__anext__ is called

__anext__ starts at i = 0

it blocks if i <= len(arr)

when a item is added (primitive, start array, start map)
arr is extended to len(arr) + 1
__anext__ must be resumed to return a AsyncJsonNode((*path, i))


{
    "seq": true/false
    "future": <fut>
    "val": val
}

"""


@define
class UNSET:
    pass


# good to split pathstate and seqstate. the overloading and if conditionals is a threat to sanity
@define
class PathState:
    future = field(factory=lambda: asyncio.get_running_loop().create_future())
    sealed: bool = field(default=False)
    updated = field(factory=lambda: asyncio.Event())
    val: Any = field(default=UNSET)


Path = tuple[str | int, ...]


class PathStore:
    def __init__(self):
        self._nodes: dict[Path, PathState] = defaultdict(PathState)

    def get(self, path: Path) -> PathState:
        return self._nodes[path]

    def set(self, path: Path, value: Any) -> None:
        state = self._nodes[path]
        # this case is not possible, but it's llm and we are parsing partial json
        if state.future.done():
            raise ValueError(f"Path {path} already has a value: {state.val}")
        else:
            state.val = value
            state.sealed = True
            self._setclear(state.updated)
            state.future.set_result(value)

    def _setclear(self, e: asyncio.Event):
        e.set(), e.clear()

    def begin_item(self, path: Path) -> None:
        state = self._nodes[path]
        if state.val == UNSET:
            state.val = []
        state.val += [UNSET]
        self._setclear(state.updated)

    def setitem(self, path: Path, index: int, value: Any) -> None:
        pass



from .store import UNSET, PathStore


class AsyncJsonNode:
    def __init__(self, path: tuple[str | int, ...], store: PathStore):
        self._path = path
        self._store = store

    def __getitem__(self, key: str | int) -> "AsyncJsonNode":
        return AsyncJsonNode(path=(*self._path, key), store=self._store)

    def __await__(self):
        return self._store.get(self._path).future.__await__()

    def __aiter__(self):
        return Cursor(self._path, self._store)

    __str__ = __repr__ = lambda self: f"AsyncJN({self.path})"


class Cursor:
    def __init__(self, path, store: PathStore) -> None:
        self._index = 0
        self._path = path
        self._store = store

    async def __anext__(self):
        state = self._store.get(self._path)
        while True:
            while state.val == UNSET:
                await state.updated.wait()
            curr = self._index
            N = len(state.val)
            consumed = curr == N
            if state.sealed and consumed:
                raise StopAsyncIteration
            if curr < N:
                self._index += 1
                return AsyncJsonNode((*self._path, curr), self._store)
            else:
                await state.updated.wait()
